/*
 * Copyright (C) 1996,1997,1998
 *	SHUDO Kazuyuki
 *
 * Permission to use, copy, modify and distribute this software
 * and its documentation for non-commercial purposes and without
 * fee is hereby granted provided this copyright notice.
 *
 * $Id$
 */


#ifndef _COMPILER_H_
#define _COMPILER_H_


#include "x86tsc.h"

#include "opcodes.h"
#define opc_invokeignored_nocheck	230

#define opc_getstatic2	231
#define opc_putstatic2	232
#define opc_getfield2	233
#define opc_putfield2	234

#define opc_methodhead	235
#define opc_exception_handler	236
#define opc_methodtail	237

#define opc_fill_cache	238
#define opc_array_check	239

#define opc_iastore1	240
#define opc_lastore1	241

#define opc_stateto0	242
#define opc_stateto1	243
#define opc_stateto2	244
#define opc_stateto3	245
#define opc_stateto4	246

#define opc_goto_st0	247
#define opc_goto_st1	248
#define opc_goto_st2	249
#define opc_goto_st3	250
#define opc_goto_st4	251

#define NOPCODES	256


/* offerred by Sun */
#include "native.h"


/* Additional type definition */

/* for Linux, and SunOS 5.5.1 or earlier */
#if defined(linux) \
  || (defined(sun) && defined(__svr4__) && !defined(_ILP32) && !defined(_LP64))
#ifndef _UINT16_T
#define _UINT16_T
typedef unsigned short uint16_t;
#endif	/* _UINT16_T */

#ifndef __BIT_TYPES_DEFINED__
#ifndef _INT16_T
#define _INT16_T
typedef short int16_t;
#endif	/* _INT16_T */
#endif	/* __BIT_TYPES_DEFINED__ */
#endif	/* OS */


/* Additional macro definition */
#define handleToClassClass(h)	(java_lang_Object_getClass((HObject *)(h)))


/*
 * OS dependent macro definition
 */
#  undef RESOLVE_SYMBOL_ON_CODE
#if defined (linux)
#  define EXECUTEJAVA_IN_ASM
#elif defined (__FreeBSD__)
#  undef  EXECUTEJAVA_IN_ASM
#else
#  undef  EXECUTEJAVA_IN_ASM
#endif


/*
 * Select appropriate malloc library
 */
# define sysMalloc	malloc
# define sysFree	free
# define sysCalloc	calloc
# define sysRealloc	realloc


/*
 * Macros
 */

#define CREDIT " shuJIT  for Sun JVM/IA32  Copyright (c) 1998 SHUDO Kazuyuki\n"

#define COMPILER_VERSION	5

#define CODESIZE_FNAME	"jit_codesize"

#ifndef MIN
# define MIN(a,b)	((a)<(b)?(a):(b))
#endif
#ifndef MAX
# define MAX(a,b)	((a)>(b)?(a):(b))
#endif

/* STR(macro) is permuted to "value of the macro" */
#define _STR(T)	#T
#define STR(MACRO)	_STR(MACRO)


/*
 * Type definition
 */

typedef struct compiled_code_info {
  int code_size;
  int ret_size;
  char *argsizes;
} CodeInfo;


typedef struct pcentry {
  unsigned short opcode;
  unsigned short state;
  unsigned int byteoff;
  unsigned int nativeoff;
} pcentry;

typedef struct jpentry {
  unsigned int tgtoff;
  unsigned int argoff;
} jpentry;

typedef struct compiler_context {
  ExecEnv *ee;
  struct methodblock *mb;

  unsigned int exc_handler_nativeoff;
  unsigned int finish_return_nativeoff;

  /* buffer for compiled code */
#define DEFAULT_BUF_SIZE	8192
  unsigned char *buffer;
  int buf_size;
  unsigned char *bufp;
  int lastbufp_off;

  /* program counter table */
#define DEFAULT_PCTABLE_SIZE	100
  int pctablesize;
  int pctablelen;
  pcentry *pctable;

  /* jump instruction table */
#define DEFAULT_JPTABLE_SIZE	100
  int jptablesize;
  int jptablelen;
  jpentry *jptable;
} CompilerContext;


/*
 * Global Variables
 */

/* in compiler.c */
extern bool_t compiler_enabled;

extern int opt_quiet;
	/* suppress initial message and some outputs */
extern int opt_outcode;
	/* write generated code to code_<classname>_<methodname>.s */
extern int opt_codesize;
	/* write code size of each methods to the file jit_codesize */
extern int opt_dontcmplvmcls;
	/* suppress compilation classes
	   which is already loaded when JIT is initialized */

extern void *sym_compileAndInvokeMethod;
extern void *sym_invokeCompiledMethod;
extern void *sym_invokeJavaMethod;
extern void *sym_invokeSynchronizedJavaMethod;


/*
 * Global Functions
 */
/* in code.c */
extern volatile void compiledCode(
	JHandle *o, struct methodblock *mb, int args_size, ExecEnv *ee
#ifdef RUNTIME_DEBUG
	,int runtime_debug
#endif
);

/* in computil.c */
extern void *access2invoker(int access);

extern void showCompilerContext(CompilerContext *cc);
extern CompilerContext *newCompilerContext(struct methodblock *mb);
extern void freeCompilerContext(CompilerContext *cc);
extern void resetCompilerContext(CompilerContext *cc);

extern void writeToBuffer(CompilerContext *cc, void *, size_t);
#define lastBufend(cc)	((cc)->buffer + (cc)->lastbufp_off)
extern void cancelOnBuffer(CompilerContext *cc, size_t);
extern void ensureBufferSize(CompilerContext *cc, size_t req);

extern void pctableAdd(CompilerContext *cc,
	int opcode, int state, unsigned int byteoff, unsigned int nativeoff);
extern void pctableInsert(CompilerContext *cc, int index,
	int opcode, int state, unsigned int byteoff, unsigned int nativeoff);
extern void pctableDelete(CompilerContext *cc, int index);
extern pcentry *pctableGet(CompilerContext *cc, int byteoff);

extern void jptableAdd(CompilerContext *cc,
	unsigned int tgtoff, unsigned int argoff);


/* in invoker.c */
extern bool_t compileAndInvokeMethod(
	JHandle *o, struct methodblock *mb, int args_size, ExecEnv *ee);
extern bool_t invokeCompiledMethod(
	JHandle *o, struct methodblock *mb, int args_size, ExecEnv *ee);

/* in compile.c */
extern int compileMethod(struct methodblock *mb);
extern void freeMethod(struct methodblock *mb);
#ifdef RESOLVE_SYMBOL_ON_CODE
void initFunctionSymbols();
#endif

/* in runtime.c */
extern int invokeMethod(
#ifdef RUNTIME_DEBUG
	int runtime_debug,
#endif
	ExecEnv *ee, int narg, int retsize,
	JHandle *obj, struct methodblock *method,
	int bytepcoff, stack_item *stackpointer);
extern struct methodblock *getInterfaceMethod(
#ifdef RUNTIME_DEBUG
	int runtime_debug,
#endif
	ExecEnv *ee,
	JHandle *obj, struct methodblock *imethod, unsigned char *guessptr);
JHandle *multianewarray(
#ifdef RUNTIME_DEBUG
	int runtime_debug,
#endif
	ExecEnv *ee, int dimensions, ClassClass *arrayclazz,
	stack_item *stackpointer);
extern struct CatchFrame *searchCatchFrame(ExecEnv *ee, struct methodblock *mb,
	int bytepcoff
#ifdef RUNTIME_DEBUG
	, int runtime_debug
#endif
);
extern CodeInfo *prepareCompiledCodeInfo(ExecEnv *ee, struct methodblock *mb);
void freeCompiledCodeInfo(CodeInfo *info);
extern void showStackFrames(ExecEnv *ee);
#ifdef RUNTIME_DEBUG
extern void showExcStackTrace(JHandle *throwable);
extern char *showObjectBody(char *sig, JHandle *obj);
#endif

#endif /* _COMPILER_H_ */
